///////////////////////////////////////////////////////////////////////////////
//
// File:    lcdshift.c
// Author:  Roger Rowland
//
// Created: November 2014
//
///////////////////////////////////////////////////////////////////////////////

#include <xc.h>
#include "global.h"
#include "lcdshift.h"

///////////////////////////////////////////////////////////////////////////////
// lcd commands
///////////////////////////////////////////////////////////////////////////////

#define LCD_CLEARDISPLAY 0x01
#define LCD_RETURNHOME 0x02
#define LCD_ENTRYMODESET 0x04
#define LCD_DISPLAYCONTROL 0x08
#define LCD_CURSORSHIFT 0x10
#define LCD_FUNCTIONSET 0x20
#define LCD_SETCGRAMADDR 0x40
#define LCD_SETDDRAMADDR 0x80

///////////////////////////////////////////////////////////////////////////////
// flags for display entry mode
///////////////////////////////////////////////////////////////////////////////

#define LCD_ENTRYRIGHT 0x00
#define LCD_ENTRYLEFT 0x02
#define LCD_ENTRYSHIFTINCREMENT 0x01
#define LCD_ENTRYSHIFTDECREMENT 0x00

///////////////////////////////////////////////////////////////////////////////
// flags for display on/off control
///////////////////////////////////////////////////////////////////////////////

#define LCD_DISPLAYON 0x04
#define LCD_DISPLAYOFF 0x00
#define LCD_CURSORON 0x02
#define LCD_CURSOROFF 0x00
#define LCD_BLINKON 0x01
#define LCD_BLINKOFF 0x00

///////////////////////////////////////////////////////////////////////////////
// flags for display/cursor shift
///////////////////////////////////////////////////////////////////////////////

#define LCD_DISPLAYMOVE 0x08
#define LCD_CURSORMOVE 0x00
#define LCD_MOVERIGHT 0x04
#define LCD_MOVELEFT 0x00

///////////////////////////////////////////////////////////////////////////////
// flags for function set
///////////////////////////////////////////////////////////////////////////////

#define LCD_8BITMODE 0x10
#define LCD_4BITMODE 0x00
#define LCD_2LINE 0x08
#define LCD_1LINE 0x00
#define LCD_5x10DOTS 0x04
#define LCD_5x8DOTS 0x00

///////////////////////////////////////////////////////////////////////////////
// local data
///////////////////////////////////////////////////////////////////////////////

// hex character map
static const char g_szHex[16] = {'0','1','2','3','4','5','6','7',
                                 '8','9','A','B','C','D','E','F'};

///////////////////////////////////////////////////////////////////////////////
// internally used function prototypes and macros
///////////////////////////////////////////////////////////////////////////////

void LCDSHRPulseEnable();                           // pulse the Enable line
void LCDSHRNibble(uint8_t uByte, uint8_t uType);    // send a command or data nibble
void LCDSHRByte(uint8_t uByte, uint8_t uType);      // send a command or data byte

// macros
#define LCDSHRCmd(c)  (LCDSHRByte(c, 0))            // send command
#define LCDSHRData(d) (LCDSHRByte(d, 1))            // send data

///////////////////////////////////////////////////////////////////////////////
// LCD function implementations
///////////////////////////////////////////////////////////////////////////////

void LCDSHRInit()
{
    // initialise LCD, so, after power on, wait for LCD to initialise
    __delay_ms(30);

    // control lines are outputs
    LCD_E_TRIS = 0;
    SHR_SI_TRIS = 0;
    SHR_SCK_TRIS = 0;

    // pull controls low
    LCD_E = 0;
    SHR_SI = 0;
    SHR_SCK = 0;

    // this is according to the hitachi HD44780 datasheet
    // page 46 figure 24

    // we start in 8-bit mode, try to set 4-bit mode
    LCDSHRNibble(0x03, 0);
    __delay_ms(5); // wait more than 4.1ms

    // second try
    LCDSHRNibble(0x03, 0);
    __delay_ms(5); // wait more than 4.1ms

    // third and final try
    LCDSHRNibble(0x03, 0);
    __delay_us(150);

    // finally, set to 4-bit interface
    LCDSHRNibble(0x02, 0);

    // set # lines, font size, etc.
    LCDSHRCmd(LCD_FUNCTIONSET | LCD_4BITMODE | LCD_2LINE | LCD_5x8DOTS);

    // turn the display on with no cursor or blinking default
    LCDSHRCmd(LCD_DISPLAYCONTROL | LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF);

    // clear the display
    LCDSHRClear();

    // set the entry mode
    LCDSHRCmd(LCD_ENTRYMODESET | LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT);
}

void LCDSHRClear()
{
    // clear LCD display
    LCDSHRCmd(LCD_CLEARDISPLAY);
    __delay_ms(2);
}

void LCDSHRWriteHex(uint8_t uValue)
{
    // write hex value to LCD
    char szHex[5] = "0x00";

    // convert value to hex
    szHex[3] = g_szHex[uValue & 0x0F];
    szHex[2] = g_szHex[(uValue >> 4) & 0x0F];

    // write as string
    LCDSHRWriteString(szHex);
}

void LCDSHRWriteString(const char *szMsg)
{
    // write string to LCD
    while (*szMsg != '\0')
    {
        LCDSHRData(*szMsg);
        ++szMsg;
    }
}

void LCDSHRGotoXY(uint8_t x, uint8_t y)
{
    // position cursor to xy
    if (y == 1) x |= 0b01000000;
    x |= 0b10000000;
    LCDSHRCmd(x);
}

void LCDSHRWrite(uint8_t uValue)
{
    // write 8-bit data via shift register
    uint8_t uMask = 0b10000000;

    // shift in 8 bits, msb first
    for (uint8_t i = 0; i < 8; ++i)
    {
        // set data and clock it in
        SHR_SI = (uValue & uMask) ? 1 : 0;
        SHR_SCK = 1;
        SHR_SCK = 0;

        // prepare for next bit
        uMask >>= 1;
    }

    // now clock out and pulse Enable
    LCDSHRPulseEnable();
}

void LCDSHRPulseEnable()
{
    // pulse the Enable line, which also clocks
    // the RCK line to the shift register to put
    // the 8-bits we've shifted to the output
    LCD_E = 0;
    __delay_us(1);
    LCD_E = 1;
    __delay_us(1);          // enable pulse must be > 450ns
    LCD_E = 0;
    __delay_us(100);        // commands need > 37us to settle
}

void LCDSHRNibble(uint8_t uByte, uint8_t uType)
{
    // send a command or data nibble
    uByte <<= 4;
    if (uType != 0) uByte |= 0x8;
    LCDSHRWrite(uByte);
}

void LCDSHRByte(uint8_t uByte, uint8_t uType)
{
    // send a command (type 0) or data byte (type 1)
    // so, send high nibble then low nibble
    LCDSHRNibble(uByte >> 4, uType);
    LCDSHRNibble(uByte & 0x0F, uType);
}
