///////////////////////////////////////////////////////////////////////////////
//
// File:    ledmatrix.c
// Author:  Roger Rowland
//
// Created: November 2014
//
///////////////////////////////////////////////////////////////////////////////

#include <xc.h>
#include "global.h"
#include "ledmatrix.h"

///////////////////////////////////////////////////////////////////////////////
// MAX7219 register addresses
///////////////////////////////////////////////////////////////////////////////

#define MAX_NOOP 0x00
#define MAX_DIG0 0x01
#define MAX_DIG1 0x02
#define MAX_DIG2 0x03
#define MAX_DIG3 0x04
#define MAX_DIG4 0x05
#define MAX_DIG5 0x06
#define MAX_DIG6 0x07
#define MAX_DIG7 0x08
#define MAX_DECODEMODE 0x09
#define MAX_INTENSITY 0x0A
#define MAX_SCANLIMIT 0x0B
#define MAX_SHUTDOWN 0x0C
#define MAX_DISPLAYTEST 0x0F

///////////////////////////////////////////////////////////////////////////////
// internally used function prototypes
///////////////////////////////////////////////////////////////////////////////

void LEDWriteByte(uint8_t uByte);                   // write given byte
void LEDWriteAll(uint8_t uAddress, uint8_t uByte);  // write given byte to address for all MAX7219's

///////////////////////////////////////////////////////////////////////////////
// local data
///////////////////////////////////////////////////////////////////////////////

uint16_t g_pRow[16] = { 0 };                        // bit map of LEDs per row

///////////////////////////////////////////////////////////////////////////////
// LED matrix function implementations
///////////////////////////////////////////////////////////////////////////////

void LEDInit()
{
    // initialise LED matrices
    LED_CLK_TRIS = 0;
    LED_DIN_TRIS = 0;
    LED_CS_TRIS = 0;

    // clear control lines
    LED_CLK = 0;
    LED_DIN = 0;
    LED_CS = 0;

    // configure all MAX7219's for individual LED access
    LEDWriteAll(MAX_DECODEMODE, 0x00);             // decoding   : none
    LEDWriteAll(MAX_INTENSITY, 0x0B);              // intensity  : 23/32
    LEDWriteAll(MAX_SCANLIMIT, 0x07);              // scan limit : 8 digits (all LEDs)
    LEDWriteAll(MAX_SHUTDOWN, 0x01);               // shutdown   : normal
    LEDWriteAll(MAX_DISPLAYTEST, 0x00);            // test mode  : off

    // ensure all LEDs are off
    LEDClear(1);
}

void LEDWriteByte(uint8_t uByte)
{
    // write given byte
    for (uint8_t i = 0; i < 8; ++i)
    {
        // clock in one bit
        LED_CLK = 0;
        LED_DIN = (uByte & 0x80) ? 1 : 0;

        // prepare for next bit
        uByte <<= 1;
        LED_CLK = 1;
    }
}

void LEDWriteAll(uint8_t uAddress, uint8_t uByte)
{
    // write given byte to address for all four MAX7219's
    LED_CS = 0;
    for (uint8_t i = 0; i < 4; ++i)
    {
        // echo to all 4 matrices
        LEDWriteByte(uAddress);
        LEDWriteByte(uByte);
    }
    LED_CS = 1;
}

void LEDClear(uint8_t bRefresh)
{
    // turn all LEDs off
    for (uint8_t i = 0; i < 16; ++i)
        g_pRow[i] = (uint16_t)0;

    // refresh if requested
    if (bRefresh) LEDRefresh();
}

void LEDSetRow(uint8_t uRow, uint16_t uData, uint8_t bRefresh)
{
    // set all 16 LEDs at given row (zero based)
    g_pRow[uRow] = uData;
    if (bRefresh) LEDRefreshRow(uRow);
}

void LEDSet(uint8_t uRow, uint8_t uCol, uint8_t bOn, uint8_t bRefresh)
{
    // set a single LED on or off
    if (bOn)
    {
        // set led on
        g_pRow[uRow] |= (0x8000 >> uCol);
    }
    else
    {
        // set led off
        g_pRow[uRow] &= ~(0x8000 >> uCol);
    }

    // refresh the row if requested
    if (bRefresh) LEDRefreshRow(uRow);
}

void LEDRefreshRow(uint8_t uRow)
{
    // refresh given row (zero based)
    t_short uData;
    uint8_t uAddress = uRow % 8;

    // start writing
    LED_CS = 0;

    // write 4 addresses (so we actually
    // cover two rows instead of one)
    uData.uShort = g_pRow[uAddress + 8];
    LEDWriteByte(uAddress + 1);
    LEDWriteByte(uData.uLSB);
    LEDWriteByte(uAddress + 1);
    LEDWriteByte(uData.uMSB);
    uData.uShort = g_pRow[uAddress];
    LEDWriteByte(uAddress + 1);
    LEDWriteByte(uData.uLSB);
    LEDWriteByte(uAddress + 1);
    LEDWriteByte(uData.uMSB);

    // latch data
    LED_CS = 1;
}

void LEDRefresh()
{
    // refresh all rows (nb - only need to do
    // 8 rows because the other 8 are included)
    for (uint8_t i = 0; i < 8; ++i)
    {
        // set state from cache
        LEDRefreshRow(i);
    }
}
