///////////////////////////////////////////////////////////////////////////////
//
// File:    main.c
// Author:  Roger Rowland
//
// Created: November 2014
//
///////////////////////////////////////////////////////////////////////////////

#include <xc.h>

// CONFIG must appear early
#pragma config FOSC = HS        // Oscillator Selection bits (HS oscillator: High-speed crystal/resonator on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled)
#pragma config PWRTE = ON       // Power-up Timer Enable bit (PWRT enabled)
#pragma config MCLRE = OFF      // RA5/MCLR/VPP Pin Function Select bit (RA5/MCLR/VPP pin function is digital input, MCLR internally tied to VDD)
#pragma config BOREN = OFF      // Brown-out Detect Enable bit (BOD disabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable bit (RB4/PGM pin has digital I/O function, HV on MCLR must be used for programming)
#pragma config CPD = OFF        // Data EE Memory Code Protection bit (Data memory code protection off)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit (Code protection off)

// include dependencies
#include "global.h"
#include "lcdshift.h"
#include "ledmatrix.h"
#include "ps2mouse.h"

// hardcoded bitmask for a 16x16 smiley
static const uint16_t g_pSmile[16] =
{
    0xFFFF, 0xF81F, 0xE007, 0xC003,
    0xC423, 0x8E71, 0x8E71, 0x8421,
    0x8001, 0xA005, 0x9FF9, 0xC813,
    0xC663, 0xE187, 0xF81F, 0xFFFF
};

void Sequence()
{
    // make some animation
    for (int8_t i = 0; i < 8; ++i)
    {
        // calculate bounds
        int8_t nStart = i, nStop = 15 - i, bOnOff;

        // set the state of all LEDs
        for (int8_t r = 0; r < 16; ++r)
        {
            for (int8_t c = 0; c < 16; ++c)
            {
                // decide if on first
                bOnOff = (r == nStart || c == nStart || r == nStop || c == nStop);

                // now turn off if outside shell
                if (r < nStart || c < nStart || r > nStop || c > nStop)
                    bOnOff = 0;

                // set LED on/off (no refresh yet)
                LEDSet(r, c, bOnOff, 0);
            }
        }

        // refresh and hold
        LEDRefresh();
        __delay_ms(100);
    }

    // reset
    LEDClear(1);
}

///////////////////////////////////////////////////////////////////////////////
// main program
///////////////////////////////////////////////////////////////////////////////

void main(void)
{
    // entry point
    CMCON = 0x7; // turn off comparators

    // enable weak pull-ups on portb and make used pins
    // input (high impedance). Note that the mouse data
    // and clock are open-collector so be careful with portb
    OPTION_REGbits.nRBPU = 0;
    TRISB = 0b00100011;             // nb - debug pins are RB2 and RB3 (output)

    // initialise the LCD module
    LCDSHRInit();

    // write status to LCD
    LCDSHRClear();
    LCDSHRWriteString("Initialising...");

    // initialise the LED matrix
    LEDInit();

    // short animation while waiting for mouse to do BAT
    for (uint8_t i = 0; i < 5; ++i)
        Sequence();

    // while the mouse is not yet initialised ...
    while (PS2GetInitStatus() != 0)
    {
        // initialise the ps/2 mouse
        uint8_t uByte = PS2Init();
        if (uByte != 0)
        {
            // failed - show status
            LCDSHRGotoXY(0, 1);
            LCDSHRWriteString("Status:");
            LCDSHRGotoXY(8, 1);
            LCDSHRWriteHex(uByte);

            // wait for retry
            __delay_ms(1000);
        }
        else
        {
            // all done - show ready
            LCDSHRClear();
            LCDSHRWriteString("Ready");

            // show the smiley
            for (uint8_t i = 0; i < 16; ++i)
                LEDSetRow(i, g_pSmile[i], 0);
            LEDRefresh();

            // short hold
            __delay_ms(3000);

            // clear the LED display
            LEDClear(1);
        }
    }

    // assume initial mouse position
    uint8_t nLastMouseX = 0, nLastMouseY = 0;
    LEDSet(15 - nLastMouseY, nLastMouseX, 1, 1);

    // start streaming mouse data
    PS2StartStreaming();

    // the mouse is initialised now, so we can
    // configure and enable interrupts to get
    // the clock signal from the mouse when it's
    // in streaming mode and update the position
    OPTION_REGbits.INTEDG = 0;                      // interrupt on falling edge
    INTCONbits.INTF = 0;                            // reset external interrupt flag
    INTCONbits.INTE = 1;                            // enable external interrupt

    // enable interrupts globally
    ei();

    // main loop
    while (1) 
    {
        // get latest mouse position
        uint8_t nX = g_uPS2PosX, nY = g_uPS2PosY;

        // if it's changed ...
        if (nX != nLastMouseX || nY != nLastMouseY)
        {
            // ... show the new position
            LEDSet(15 - nY, nX, 1, 1);

            // if not drawing ...
            if (!(g_uPSBtn & 0x1))
            {
                // ... erase last position
                LEDSet(15 - nLastMouseY, nLastMouseX, 0, 1);
            }

            // update current position
            nLastMouseX = nX;
            nLastMouseY = nY;
        }

        // if right mouse button clicked ...
        if (g_uPSBtn & 0x02)
        {
            // ... clear display
            LEDClear(1);
        }
    }
}

void interrupt isr()
{
    // interupt service routine - only
    // one interrupt is ever enabled
    // but check anyway ...
    if (INTCONbits.INTF)
    {
        // delegate processing to mouse
        PS2Interrupt();

        // reset external interrupt flag
        INTCONbits.INTF = 0;
    }
}