///////////////////////////////////////////////////////////////////////////////
//
// File:    ps2mouse.c
// Author:  Roger Rowland
//
// Created: November 2014
//
///////////////////////////////////////////////////////////////////////////////

#include <xc.h>
#include "global.h"
#include "ps2mouse.h"

///////////////////////////////////////////////////////////////////////////////
// PS/2 mouse commands
///////////////////////////////////////////////////////////////////////////////

#define PS2_RESET 0xFF
#define PS2_RESEND 0xFE
#define PS2_SET_DEFAULTS 0xF6
#define PS2_DISABLE_DATA 0xF5
#define PS2_ENABLE_DATA 0xF4
#define PS2_SAMPLE_RATE 0xF3
#define PS2_DEVICE_ID 0xF2
#define PS2_REMOTE 0xF0
#define PS2_SET_WRAP 0xEE
#define PS2_RESET_WRAP 0xEC
#define PS2_READ_DATA 0xEB
#define PS2_STREAM 0xEA
#define PS2_STATUS 0xE9

///////////////////////////////////////////////////////////////////////////////
// PS/2 mouse return codes
///////////////////////////////////////////////////////////////////////////////

#define PS2_ACK 0xFA
#define PS2_ERROR 0xFC
#define PS2_TEST_OK 0xAA

///////////////////////////////////////////////////////////////////////////////
// short and long timeout defaults and other local data
///////////////////////////////////////////////////////////////////////////////

// constants
static const uint16_t PS2_SHORT_TIMEOUT = (uint16_t)(_XTAL_FREQ / 4000.0);  // 1 ms
static const uint32_t PS2_LONG_TIMEOUT = (uint32_t)(_XTAL_FREQ / 80.0);     // 50 ms

// local variables
volatile t_byte g_pPacket[3];                           // movement data packet
volatile uint8_t g_uBit, g_uByte;                       // packet bit and byte indices
t_byte g_pInitStatus = 0xDD;                            // initialisation status
t_byte g_pShadow = 0x00;                                // shadow port
uint8_t g_pTimeout;                                     // timeout flag

///////////////////////////////////////////////////////////////////////////////
// macros for avoiding RMW problems on PS/2 port
///////////////////////////////////////////////////////////////////////////////

// pull clock line low
#define PS2ClockLow() { g_pShadow.PS2_CLK_BIT = 0; PS2_CLK_TRIS = 0; PS2_PORT = g_pShadow.uByte; }

// release clock line but drive it high first just get a faster rise time
#define PS2ClockRelease() { g_pShadow.PS2_CLK_BIT = 1; PS2_PORT = g_pShadow.uByte; PS2_CLK_TRIS = 1; }

// pull data line low
#define PS2DataLow() { g_pShadow.PS2_DAT_BIT = 0; PS2_DAT_TRIS = 0; PS2_PORT = g_pShadow.uByte; }

// release data line but drive it high first just get a faster rise time
#define PS2DataRelease() { g_pShadow.PS2_DAT_BIT = 1; PS2_PORT = g_pShadow.uByte; PS2_DAT_TRIS = 1; }

// set a debug output on the PS/2 port
#define PS2Debug(p, n) { g_pShadow.p = n; PS2_PORT = g_pShadow.uByte; }

///////////////////////////////////////////////////////////////////////////////
// internally used function prototypes
///////////////////////////////////////////////////////////////////////////////

uint8_t PS2Write(t_byte uByte);                         // write byte and read ack
uint8_t PS2WriteByte(t_byte uByte);                     // write cmd or data byte
uint8_t PS2ReadByte(t_byte *uByte);                     // read one byte from mouse
uint8_t PS2WaitLongClockLow();                          // wait for mouse to pull clock low (long timeout)
uint8_t PS2WaitClockLow();                              // wait for mouse to pull clock low
uint8_t PS2WaitClockHigh();                             // wait for mouse to release clock
uint8_t PS2WaitClockHighLow();                          // wait for mouse to release clock then pull low
uint8_t PS2WaitDataLow();                               // wait for mouse to pull data low
uint8_t PS2WaitDataHigh();                              // wait for mouse to release data

///////////////////////////////////////////////////////////////////////////////
// PS/2 mouse function implementations
///////////////////////////////////////////////////////////////////////////////

uint8_t PS2Init()
{
    // initialise ps/2 mouse, returning zero if
    // ok but setting error flags as follows if not:
    //
    //    7 6 5 4 3 2 1 0
    //    - - - - - - - -
    //    0 A B C D E F G
    //
    // A : 1 means PS2WriteByte timeout during PS2Write reset cmd
    // B : 1 means PS2ReadByte timeout during PS2Write reset cmd
    // C : 1 means no ACK received during PS2Write reset cmd
    // D : 1 means PS2ReadByte timeout reading self-test result
    // E : 1 means self-test result returned wrong code
    // F : 1 means PS2ReadByte timeout reading device id
    // G : 1 means device id returned wrong code
    t_byte uByte;

    // send reset command
    uByte.uByte = PS2_RESET;
    g_pInitStatus.uByte = PS2Write(uByte);

    // allow time for self-test
    __delay_ms(10);

    // now we expect a success code
    g_pInitStatus.b3 = PS2ReadByte(&uByte);
    if (uByte.uByte != PS2_TEST_OK)
        g_pInitStatus.b2 = 1;

    // check device id
    g_pInitStatus.b1 = PS2ReadByte(&uByte);
    if (uByte.uByte != 0x00)
        g_pInitStatus.b0 = 1;

    // enable data reporting
    uByte.uByte = PS2_ENABLE_DATA;
    if (PS2Write(uByte) != 0)
        g_pInitStatus.b7 = 1;

    // block mouse but release data
    PS2ClockLow();
    PS2DataRelease();

    // set default position (0, 0) and no buttons pressed
    g_uPS2PosX = 0;
    g_uPS2PosY = 0;
    g_uPSBtn = 0;

    // initialise packet indices
    g_uBit = 0;
    g_uByte = 0;

    // initialise packet
    g_pPacket[0].uByte = 0;
    g_pPacket[1].uByte = 0;
    g_pPacket[2].uByte = 0;

    // return zero if we succeeded
    return g_pInitStatus.uByte;
}

uint8_t PS2GetInitStatus()
{
    // return initialisation status
    return g_pInitStatus.uByte;
}

inline void PS2StartStreaming()
{
    // start ps/2 mouse streaming mode
    PS2ClockRelease();
}

uint8_t PS2Write(t_byte uByte)
{
    // write byte and read ack, returning zero if
    // ok but setting error flags as follows if not:
    //
    //    7 6 5 4 3 2 1 0
    //    - - - - - - - -
    //    0 A B C 0 0 0 0
    //
    // A : 1 means PS2WriteByte timeout
    // B : 1 means PS2ReadByte timeout
    // C : 1 means no ACK received
    t_byte uError;

    // initialise return code
    uError.uByte = 0;

    // write and read reply
    uError.b6 = PS2WriteByte(uByte);
    uError.b5 = PS2ReadByte(&uByte);

    // block clock and release data
    PS2ClockLow();
    PS2DataRelease();

    // check acknowledgement
    if (uByte.uByte != PS2_ACK)
        uError.b4 = 1;

    // return zero if we succeeded
    return uError.uByte;
}

uint8_t PS2WriteByte(t_byte uByte)
{
    // write cmd or data byte
    t_byte pParity;

    // initialise
    g_pTimeout = 0;
    pParity.uByte = 0;

    // calculate parity (odd parity required)
    for (uint8_t i = 1; i != 0; i <<= 1)
    {
        // count the zeros
        if (!(uByte.uByte & i))
            ++pParity.uByte;
    }

    // make sure we get odd parity
    pParity.uByte = (pParity.b0 == 1) ? 0 : 1;

    // request to send
    PS2ClockLow();
    __delay_us(200);
    PS2DataLow();
    PS2ClockRelease();

    // dummy loop for clean code (avoids gotos)
    do
    {
        // wait for the mouse to pull clock low (long timeout)
        if (!PS2WaitLongClockLow()) break;

        // send 8 bits of data (lsb first) and parity bit
        // (nb - loop unrolled for speed)
        PS2_DAT = uByte.b0;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b1;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b2;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b3;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b4;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b5;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b6;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = uByte.b7;
        if (!PS2WaitClockHighLow()) break;
        PS2_DAT = pParity.b0;
        if (!PS2WaitClockHighLow()) break;

        // release data line
        PS2DataRelease();

        // wait for acknowledge
        if (!PS2WaitDataLow()) break;
        if (!PS2WaitClockLow()) break;
        if (!PS2WaitDataHigh()) break;
        if (!PS2WaitClockHigh()) break;
        
        // all ok, break loop
        break;
    }
    while (1);

    // always release data (in case of earlier timeouts)
    PS2DataRelease();

    // return 1 if timeout detected
    return g_pTimeout;
}

uint8_t PS2ReadByte(t_byte *uByte)
{
    // read one byte from mouse
    uByte->uByte = 0;
    g_pTimeout = 0;

    // release clock to allow mouse to respond
    PS2ClockRelease();

    // dummy loop for clean code (avoids gotos)
    do
    {
        // wait for the mouse to pull clock low (long timeout)
        if (!PS2WaitLongClockLow()) break;

        // we expect to see 11 clock pulses but we will
        // ignore the start bit, parity bit and stop bit
        // (nb - loop unrolled for speed)
        if (!PS2WaitClockHighLow()) break;
        uByte->b0 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b1 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b2 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b3 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b4 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b5 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b6 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        uByte->b7 = PS2_DAT;
        if (!PS2WaitClockHighLow()) break;
        if (!PS2WaitClockHighLow()) break;

        // all ok - break loop
        break;
    }
    while (1);

    // return 1 if timeout detected
    return g_pTimeout;
}

inline void PS2Interrupt()
{
    // called on falling clock edge from isr
    // so we can read streaming data
    uint8_t uBit = PS2_DAT;

    // do nothing unless initialised
    if (g_pInitStatus.uByte != 0) return;

    // we expect 11 bits per byte so we ignore the
    // leading start bit and trailing parity and stop
    if (g_uBit > 0 && g_uBit < 9)
    {
        // save incoming bit into movement packet
        if (uBit)
            g_pPacket[g_uByte].uByte |= (uBit << (g_uBit - 1));
        else
            g_pPacket[g_uByte].uByte &= ~(uBit << (g_uBit - 1));
    }

    // increment bit index
    if (++g_uBit > 10)
    {
        // end of byte
        g_uBit = 0;

        // if we have a complete packet ...
        if (++g_uByte > 2)
        {
            // reset packet index
            g_uByte = 0;
            
            // analyse movement data, so first
            // zeroise any movement if overflow
            if (g_pPacket[0].b6) g_pPacket[1].uByte = 0;
            if (g_pPacket[0].b7) g_pPacket[2].uByte = 0;

            // extract deltas
            t_short nDeltaX, nDeltaY;
            nDeltaX.uLSB = g_pPacket[1].uByte;
            nDeltaX.uMSB = (g_pPacket[0].b4) ? 0xFF : 0x00;
            nDeltaY.uLSB = g_pPacket[2].uByte;
            nDeltaY.uMSB = (g_pPacket[0].b5) ? 0xFF : 0x00;

            // calculate new position
            int16_t nX = (int16_t)g_uPS2PosX + nDeltaX.nShort;
            int16_t nY = (int16_t)g_uPS2PosY + nDeltaY.nShort;

            // clamp to limits
            if (nX < 0)
                nX = 0;
            else if (nX > 15)
                nX = 15;
            if (nY < 0)
                nY = 0;
            else if (nY > 15)
                nY = 15;

            // save new position
            g_uPS2PosX = (uint8_t)nX;
            g_uPS2PosY = (uint8_t)nY;

            // save button presses
            g_uPSBtn = g_pPacket[0].uByte & 0x07;

            // reset packet
            g_pPacket[0].uByte = 0;
            g_pPacket[1].uByte = 0;
            g_pPacket[2].uByte = 0;
        }
    }
}

uint8_t PS2WaitLongClockLow()
{
    // wait for mouse to pull clock low (long timeout)
    uint32_t uTimeout = 0;
    while (PS2_CLK != 0 && uTimeout < PS2_LONG_TIMEOUT)
        ++uTimeout;

    // set flag if timeout
    if (PS2_CLK != 0) g_pTimeout = 1;

    // return true if ok
    return (g_pTimeout == 0);
}

uint8_t PS2WaitClockLow()
{
    // wait for mouse to pull clock low (short timeout)
    uint16_t uTimeout = 0;
    while (PS2_CLK != 0 && uTimeout < PS2_SHORT_TIMEOUT)
        ++uTimeout;

    // set flag if timeout
    if (PS2_CLK != 0) g_pTimeout = 1;

    // return true if ok
    return (g_pTimeout == 0);
}

uint8_t PS2WaitClockHigh()
{
    // wait for mouse to release clock (short timeout)
    uint16_t uTimeout = 0;
    while (PS2_CLK != 1 && uTimeout < PS2_SHORT_TIMEOUT)
        ++uTimeout;

    // set flag if timeout
    if (PS2_CLK != 1) g_pTimeout = 1;

    // return true if ok
    return (g_pTimeout == 0);
}

uint8_t PS2WaitClockHighLow()
{
    // wait for mouse to release clock then pull low (short timeouts)
    uint16_t uTimeout = 0;
    while (PS2_CLK != 1 && uTimeout < PS2_SHORT_TIMEOUT)
        ++uTimeout;

    // catch timeout waiting for high
    if (PS2_CLK != 1)
    {
        // timeout
        g_pTimeout = 1;
    }
    else
    {
        // now wait for low
        uTimeout = 0;
        while (PS2_CLK != 0 && uTimeout < PS2_SHORT_TIMEOUT)
            ++uTimeout;

        // set flag if timeout
        if (PS2_CLK != 0) g_pTimeout = 1;
    }

    // return true if ok
    return (g_pTimeout == 0);
}

uint8_t PS2WaitDataLow()
{
    // wait for mouse to pull data low (short timeout)
    uint16_t uTimeout = 0;
    while (PS2_DAT != 0 && uTimeout < PS2_SHORT_TIMEOUT)
        ++uTimeout;

    // set flag if timeout
    if (PS2_DAT != 0) g_pTimeout = 1;

    // return true if ok
    return (g_pTimeout == 0);
}

uint8_t PS2WaitDataHigh()
{
    // wait for mouse to release data (short timeout)
    uint16_t uTimeout = 0;
    while (PS2_DAT != 1 && uTimeout < PS2_SHORT_TIMEOUT)
        ++uTimeout;

    // set flag if timeout
    if (PS2_DAT != 1) g_pTimeout = 1;

    // return true if ok
    return (g_pTimeout == 0);
}
